import math
import numpy
import pprint
from .db import CorpsEval
from .eval import EXPECT_EARN, red as eval_red, mil as eval_mil, blue as eval_blue, growth as eval_growth

from db_hj3415 import mongo
from util_hj3415 import utils
from krx_hj3415 import krx
import copy

import logging

logger = logging.getLogger(__name__)
formatter = logging.Formatter('%(levelname)s: [%(name)s] %(message)s')
ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.setLevel(logging.WARNING)


def red(code: str) -> tuple:
    """red price와 최근 주가의 괴리율 파악

    양수면 주가가 고평가되어 있는 상태, 음수면 저평가
    음수가 현재 주가가 싸다는 의미

    Returns:
        tuple : 괴리율을 기반으로한 포인트, 주가와 red price 비교한 괴리율

    Note:
        << eval.red() 값 >> \n
        {'red_price': 47896, '사업가치': 7127.5, '재산가치': 7152.78, '부채평가': 902.4, '발행주식수': 27931.0}\n
    """
    if utils.is_6digit(code):
        c101 = mongo.C101(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    logger.debug(f'c101 {c101.get_recent()}')

    try:
        recent_price = utils.to_int(c101.get_recent()['주가'])
    except KeyError:
        recent_price = float('nan')
    red_price = eval_red(code=code)['red_price']

    logger.info(f"recent_price : {recent_price}\tred_price : {red_price}")

    try:
        # 괴리율 구하는 공식이 두가지임. 어떤걸 사용해도 동일함
        f1 = round((recent_price / red_price - 1) * 100, 2)
        f2 = round((recent_price - red_price) / red_price * 100, 2)
        logger.debug(f'f1 : {f1}, f2 : {f2}')
        괴리율 = round((recent_price / red_price - 1) * 100, 2)
    except ZeroDivisionError:
        괴리율 = float('nan')

    logger.info(f'괴리율 : {괴리율}')

    if math.isnan(괴리율) or red_price <= 0:
        return 0, float('nan')
    else:
        try:
            p = round(100*math.log10(-괴리율+31.622777)-150)
            return p if p > 0 else 0, 괴리율
        except ValueError:
            # 괴리율+31.622777이 0이하인 경우 ValueError 발생함.
            return 0, 괴리율


def mil(code: str) -> tuple:
    """
    - 재무활동현금흐름이 마이너스라는 것은 배당급 지급했거나, 자사주 매입했거나, 부채를 상환한 상태임.
    - 반대는 채권자로 자금을 조달했거나 신주를 발행했다는 의미
    <주주수익률> - 재무활동현금흐름/시가총액 => 5%이상인가?

    투하자본수익률(ROIC)가 30%이상인가
    ROE(자기자본이익률) 20%이상이면 아주 우수 다른 투자이익률과 비교해볼것 10%미만이면 별로...단, 부채비율을 확인해야함.

    이익지표 ...영업현금흐름이 순이익보다 많은가 - 결과값이 음수인가..

    FCF는 영업현금흐름에서 자본적 지출(유·무형투자 비용)을 차감한 순수한 현금력이라 할 수 있다.
    말 그대로 자유롭게(Free) 사용할 수 있는 여윳돈을 뜻한다.
    잉여현금흐름이 플러스라면 미래의 투자나 채무상환에 쓸 재원이 늘어난 것이다.
    CAPEX(Capital expenditures)는 미래의 이윤을 창출하기 위해 지출된 비용을 말한다.
    이는 기업이 고정자산을 구매하거나, 유효수명이 당회계년도를 초과하는 기존의 고정자산에 대한 투자에 돈이 사용될 때 발생한다.

    잉여현금흐름이 마이너스일때는 설비투자가 많은 시기라 주가가 약세이며 이후 설비투자 마무리되면서 주가가 상승할수 있다.
    주가는 잉여현금흐름이 증가할때 상승하는 경향이 있다.
    fcf = 영업현금흐름 - capex

    가치지표평가
    price to fcf 계산
    https://www.investopedia.com/terms/p/pricetofreecashflow.asp
    pcr보다 정확하게 주식의 가치를 평가할수 있음. 10배이하 추천

    Returns:
        tuple: 주주수익률, 이익지표, 투자수익률, PFCF 포인트

    Note:
        << mil_dict >>
        {'date': ['2021/03'],
         '가치지표': {'FCF': {'2016/12': -15.8, '2017/12': 62.7, '2018/12': 278.1, '2019/12': -156.1, '2020/12': -170.2, '2021/12': -146.0},
                  'PCR': {'2020/03': -3.73, '2020/06': -5.11, '2020/09': -6.15, '2020/12': -6.93, '2021/03': -7.86},
                  'PFCF': {'2016/12': -156.08, '2017/12': 39.33, '2018/12': 8.87, '2019/12': -15.8, '2020/12': -14.49, '2021/12': -16.89}},
         '이익지표': 0.0792,
         '주주수익률': 1.29,
         '투자수익률': {'ROE': -100.94, 'ROIC': -480.97}}
    """

    if utils.is_6digit(code):
        corps_eval = CorpsEval(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    mil_dict = eval_mil(code)

    logger.info(pprint.pformat(mil_dict, width=200))

    # 주주수익률 평가
    if math.isnan(mil_dict['주주수익률']):
        p1 = 0
    else:
        주주수익률평가 = math.ceil(mil_dict['주주수익률'] - (EXPECT_EARN * 100))
        p1 = 0 if 0 > 주주수익률평가 else 주주수익률평가

    # 이익지표 평가
    p2 = 10 if mil_dict['이익지표'] < 0 else 0

    # 투자수익률 평가
    MAX3 = 20
    p3 = 0
    roic = mil_dict['투자수익률']['ROIC']
    roe = mil_dict['투자수익률']['ROE']
    if math.isnan(roic) or roic <= 0:
        # roic 가 비정상이라 평가할 수 없는 경우
        if 10 < roe <= 20:
            p3 += round(MAX3 * 0.333)
        elif 20 < roe:
            p3 += round(MAX3 * 0.666)
    elif 0 < roic:
        # roic 로 평가할 수 있는 경우
        if 0 < roic <= 15:
            p3 += round(MAX3 * 0.333)
        elif 15 < roic <= 30:
            p3 += round(MAX3 * 0.666)
        elif 30 < roic:
            p3 += MAX3

    # PFCF 평가
    _, pfcf = corps_eval.get_recent(mil_dict['가치지표']['PFCF'])
    logger.info(f'recent pfcf {_}, {pfcf}')
    try:
        p = round(-40 * math.log10(pfcf) + 40)
    except ValueError:
        p = 0
    p4 = 0 if 0 > p else p

    return p1, p2, p3, p4


def blue(code: str) -> tuple:
    """회사의 안정성을 보는 지표들

    0을 기준으로 상태가 좋치 않을 수록 마이너스 값을 가진다.

    Returns:
        tuple : 유동비율, 이자보상배율, 순부채비율, 순운전자본회전율, 재고자산회전율 평가 포인트

    Notes:
        <유동비율>
        100미만이면 주의하나 현금흐름창출력이 좋으면 괜찮을수 있다.
        만약 100%이하면 유동자산에 추정영업현금흐름을 더해서 다시계산해보아 기회를 준다.
        <이자보상배율>
        이자보상배율 영업이익/이자비용으로 1이면 자금사정빡빡 5이상이면 양호
        <순운전자금회전율>
        순운전자금 => 기업활동을 하기 위해 필요한 자금 (매출채권 + 재고자산 - 매입채무)
        순운전자본회전율은 매출액/순운전자본으로 일정비율이 유지되는것이 좋으며 너무 작아지면 순운전자본이 많아졌다는 의미로 재고나 외상이 쌓인다는 뜻
        <재고자산회전율>
        재고자산회전율은 매출액/재고자산으로 회전율이 낮을수록 재고가 많다는 이야기이므로 불리 전년도등과 비교해서 큰차이 발생하면 알람.
        재고자산회전율이 작아지면 재고가 쌓인다는뜻
        <순부채비율>
        부채비율은 업종마다 달라 일괄비교 어려우나 순부채 비율이 20%이하인것이 좋고 꾸준히 늘어나지 않는것이 좋다.
        순부채 비율이 30%이상이면 좋치 않다.

        {'date': ['2020/12', '2021/03'],
        '순부채비율': (-110.85, {'2016/12': nan, '2017/12': -80.06, '2018/12': -98.37, '2019/12': -106.45, '2020/12': -89.8, '2021/12': nan}),
        '순운전자본회전율': (-0.05, {'2016/12': nan, '2017/12': nan, '2018/12': nan, '2019/12': 0.0, '2020/12': -0.09, '2021/12': nan}),
        '유동비율': 116.21,
        '이자보상배율': (-4.19, {'2016/12': nan, '2017/12': -135.49, '2018/12': -251.12, '2019/12': -2.38, '2020/12': -3.44, '2021/12': nan}),
        '재고자산회전율': (nan, {'2016/12': nan, '2017/12': nan, '2018/12': nan, '2019/12': nan, '2020/12': nan, '2021/12': nan})}
    """
    def _calc_point_with_std(data: dict) -> int:
        """표준편차를 통해 점수를 계산하는 내부 함수

        Args:
            data(dict): 재무재표상의 연/분기 딕셔너리 데이터
        """
        NEG_MAX = -5
        d_values = [i for i in data.values() if not math.isnan(i)]
        logger.info(f'd_values : {d_values}')
        if len(d_values) == 0:
            p = NEG_MAX
        else:
            std = numpy.std(d_values)
            # 표준편차가 작을수록 데이터의 변환가 적다는 의미임.
            logger.info(f'표준편차 : {std}')
            p = NEG_MAX if float(std) > -NEG_MAX else -math.floor(float(std))

        return int(p)

    if utils.is_6digit(code):
        corps_eval = CorpsEval(code)
        c104 = mongo.C104(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    blue_dict = eval_blue(code)

    logger.info(pprint.pformat(blue_dict, width=200))

    def 유동비율평가(유동비율: float) -> int:
        # 채점은 0을 기준으로 마이너스 해간다. 즉 0이 제일 좋은 상태임.
        # 유동비율 평가 - 100 이하는 문제 있음
        NEG_MAX = -10
        if math.isnan(유동비율) or 유동비율 <= 0:
            p = NEG_MAX
        else:
            p = 0 if 100 < round(유동비율) else NEG_MAX + round(유동비율/10)
        logger.info(f'유동비율평가 point : {p}')
        return int(p)

    p1 = 유동비율평가(blue_dict['유동비율'])

    def 이자보상배율평가(이자보상배율: tuple) -> int:
        # 이자보상배율평가 : 1이면 자금사정 빡빡 5이상이면 양호
        NEG_MAX = -5
        최근분기, dict_y = 이자보상배율

        if math.isnan(최근분기) or 최근분기 <= 1:
            # 최근 분기의 값이 비정상이면 최근 년도를 한번 더 비교해 보지만 좀더 엄격하게 전년대비도 비교한다.
            _, 최근연도 = corps_eval.get_recent(dict_y)
            c104.chg_page(page='c104y')
            전년대비 = c104.find_cmp(title='이자보상배율')
            if math.isnan(최근연도) or 최근연도 <= 1 or math.isnan(전년대비) or 전년대비 < 0:
                p = NEG_MAX
            else:
                p = 0 if 5 < 최근연도 else NEG_MAX + round(최근연도)
        else:
            p = 0 if 5 < 최근분기 else NEG_MAX + round(최근분기)
        logger.info(f'이자보상배율평가 point : {p}')
        return int(p)

    p2 = 이자보상배율평가(blue_dict['이자보상배율'])

    def 순부채비율평가(순부채비율: tuple) -> int:
        # 부채비율은 업종마다 달라 일괄비교 어려우나 순부채 비율이 20%이하인것이 좋고 꾸준히 늘어나지 않는것이 좋다.
        # 순부채 비율이 30%이상이면 좋치 않다.
        NEG_MAX = -5
        최근분기, dict_y = 순부채비율

        if math.isnan(최근분기) or 최근분기 >= 80:
            # 최근 분기의 값이 비정상이면 최근 년도를 한번 더 비교해 보지만 좀더 엄격하게 전년대비도 비교한다.
            _, 최근연도 = corps_eval.get_recent(dict_y)
            c104.chg_page(page='c104y')
            전년대비 = c104.find_cmp(title='순부채비율')
            if math.isnan(최근연도) or 최근연도 >= 80 or math.isnan(전년대비) or 전년대비 > 0:
                p = NEG_MAX
            else:
                p = 0 if 최근연도 < 30 else round((30 - 최근연도) / 10)
        else:
            p = 0 if 최근분기 < 30 else round((30 - 최근분기) / 10)
        logger.info(f'순부채비율평가 point : {p}')
        return int(p)

    p3 = 순부채비율평가(blue_dict['순부채비율'])

    def 순운전자본회전율평가(순운전자본회전율: tuple) -> int:
        # 순운전자본회전율은 매출액/순운전자본으로 일정비율이 유지되는것이 좋으며 너무 작아지면 순운전자본이 많아졌다는 의미로 재고나 외상이 쌓인다는 뜻
        _, dict_y = 순운전자본회전율
        p = _calc_point_with_std(data=dict_y)
        logger.info(f'순운전자본회전율평가 point : {p}')
        return p

    p4 = 순운전자본회전율평가(blue_dict['순운전자본회전율'])

    def 재고자산회전율평가(재고자산회전율: tuple) -> int:
        # 재고자산회전율은 매출액/재고자산으로 회전율이 낮을수록 재고가 많다는 이야기이므로 불리 전년도등과 비교해서 큰차이 발생하면 알람.
        # 재고자산회전율이 작아지면 재고가 쌓인다는뜻
        _, dict_y = 재고자산회전율
        p = _calc_point_with_std(data=dict_y)
        logger.info(f'재고자산회전율평가 point : {p}')
        return p

    p5 = 재고자산회전율평가(blue_dict['재고자산회전율'])

    return p1, p2, p3, p4, p5


def growth(code: str) -> tuple:
    """회사의 성장성을 보는 지표들

    <매출액>
    매출액은 어떤경우에도 성장하는 기업이 좋다.매출이 20%씩 늘어나는 종목은 유망한 종목
    <영업이익률>
    영업이익률은 기업의 경쟁력척도로 경쟁사에 비해 높으면 경제적해자를 갖춘셈

    {'date': '2020/12',
    '매출액증가율': (0.36, {'2016/12': nan, '2017/12': nan, '2018/12': nan, '2019/12': nan, '2020/12': nan, '2021/12': nan}),
    '영업이익률': {'Unnamed': nan, '파멥신': '-42641.13'},
    }

    Returns:
        tuple : 매출액증가율, 영업이익률 평가 포인트
    """

    if utils.is_6digit(code):
        corps_eval = CorpsEval(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    growth_dict = eval_growth(code)

    logger.debug(pprint.pformat(growth_dict, width=200))

    def 매출액증가율평가(매출액증가율: tuple) -> int:
        # 매출액은 어떤경우에도 성장하는 기업이 좋다.매출이 20%씩 늘어나는 종목은 유망한 종목
        MAX = 20
        최근분기, dict_y = 매출액증가율
        _, 최근연도 = corps_eval.get_recent(dict_y)

        # 최근 자료가 성장하는 중인지 판단
        if math.isnan(최근분기):
            최근분기 = 최근연도

        sp1 = 0
        if math.isnan(최근연도):
            pass
        elif 0 < 최근연도 and 0 < 최근분기:
            # 최근에 마이너스 성장이 아닌경우 MAX/10점 보너스
            sp1 += MAX / 10
            if 최근연도 < 최근분기:
                # 최근에 이전보다 더 성장중이면 MAX/10점 보너스
                sp1 += MAX / 10
            # 나머지는 성장률 기반 점수 배정
            sp1 += MAX / 2 if 최근분기 > MAX else 최근분기 / 2
        elif 최근연도 <= 0 < 최근분기:
            # 직전에 마이너스였다가 최근에 회복된 경우 MAX/10점 보너스
            sp1 += MAX / 10
            # 나머지는 성장률 기반 점수 배정
            sp1 += MAX / 2 if 최근분기 > MAX else 최근분기 / 2
        else:
            # 최근 자료가 마이너스인 경우 마이너스만큼 점수를 차감한다.
            sp1 += -(MAX / 2) if 최근분기 < -MAX else 최근분기 / 2

        # 평균매출액증가율 구하기
        d_values = [i for i in dict_y.values() if not math.isnan(i)]
        logger.info(f'd_values : {d_values}')

        평균매출액증가율 = float(numpy.mean(d_values))
        logger.info(f'평균 : {평균매출액증가율}')

        sp2 = 0
        if math.isnan(평균매출액증가율):
            sp2 += -(MAX/2)
        elif 평균매출액증가율 <= 0:
            # 평균매출액증가율이 마이너스인 경우 마이너스만큼 점수를 차감한다.
            sp2 += -(MAX / 2) if 평균매출액증가율 < -MAX else 평균매출액증가율 / 2
        else:
            sp2 = MAX / 2 if 평균매출액증가율 > MAX else 평균매출액증가율 / 2

        logger.info(f'매출액증가율평가 point : {sp1 + sp2}')

        return int(sp1 + sp2)

    p1 = 매출액증가율평가(growth_dict['매출액증가율'])

    def 영업이익률평가(영업이익률: dict) -> int:
        # 영업이익률은 기업의 경쟁력척도로 경쟁사에 비해 높으면 경제적해자를 갖춘셈
        영업이익률 = copy.deepcopy(영업이익률)
        name = krx.get_name(code)

        p = 0
        try:
            myprofit = utils.to_float(영업이익률.pop(name))
        except KeyError:
            logger.warning(f'{name} 영업이익률 does not exist.')
            return 0
        logger.info(f'종목영업이익률 : {myprofit}')

        for profit in 영업이익률.values():
            profit = utils.to_float(profit)
            if math.isnan(profit):
                continue
            elif myprofit > profit:
                p += 1
            else:
                continue

        logger.info(f'영업이익률평가 point : {p}')
        return p

    p2 = 영업이익률평가(growth_dict['영업이익률'])

    return p1, p2



"""red, mil, blue 3가지 분야에서 자료를 계산하여 리턴하는 함수 모음
"""
import math

from .db import CorpsEval
from db_hj3415 import mongo, setting
from util_hj3415 import utils
from krx_hj3415 import krx
import pandas as pd
import time
from multiprocessing import Process, cpu_count, Queue
import datetime

import logging

logger = logging.getLogger(__name__)
formatter = logging.Formatter('%(levelname)s: [%(name)s] %(message)s')
ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.setLevel(logging.WARNING)

# 주식을 통한 기대수익률 - 금리가 3%일 경우 두배인 6% 정도로 잡는다.
EXPECT_EARN = 0.04


def red(code: str) -> dict:
    if utils.is_6digit(code):
        corps_eval = CorpsEval(code)
        c101 = mongo.C101(code)
        c103 = mongo.C103(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    d1, 지배주주당기순이익 = corps_eval.calc당기순이익(nan_to_zero=True)
    d2, 유동자산 = corps_eval.calc유동자산(nan_to_zero=True)
    d3, 유동부채 = corps_eval.calc유동부채(nan_to_zero=True)
    d4, 부채평가 = corps_eval.calc비유동부채(nan_to_zero=True)

    c103.chg_page('c103재무상태표q')
    d5, 투자자산 = c103.latest_value('투자자산', nan_to_zero=True)
    d6, 투자부동산 = c103.latest_value('투자부동산', nan_to_zero=True)

    # 사업가치 계산 - 지배주주지분 당기순이익 / 기대수익률
    사업가치 = round(지배주주당기순이익 / EXPECT_EARN, 2)

    # 재산가치 계산 - 유동자산 - (유동부채*1.2) + 고정자산중 투자자산
    재산가치 = round(유동자산 - (유동부채 * 1.2) + 투자자산 + 투자부동산, 2)

    발행주식수 = c103.latest_value('발행주식수')[1]
    if math.isnan(발행주식수):
        rc101 = c101.get_recent()
        logger.error(rc101)
        발행주식수 = utils.to_int(rc101.get('발행주식'))
    else:
        발행주식수 = 발행주식수 * 1000

    try:
        red_price = round(((사업가치 + 재산가치 - 부채평가) * 100000000) / 발행주식수)
    except (ZeroDivisionError, ValueError) as e:
        logger.error(f'{e} : {code}')
        red_price = float('nan')

    logger.debug(f'Red Price : {red_price}원')
    return {
        'red_price': red_price,
        '사업가치': 사업가치,
        '재산가치': 재산가치,
        '부채평가': 부채평가,
        '발행주식수': 발행주식수,
        'date': [i for i in {d1, d2, d3, d4, d5, d6} if i != ''],  # ''값을 제거하고 리스트로 바꾼다.
    }


def mil(code: str) -> dict:
    if utils.is_6digit(code):
        corps_eval = CorpsEval(code)
        c103 = mongo.C103(code)
        c104 = mongo.C104(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    marketcap = corps_eval.marketcap
    logger.info(f'{code} market cap: {marketcap}')
    fcf_dict = corps_eval.findFCF()
    pfcf_dict = corps_eval.findPFCF()
    d1, 지배주주당기순이익 = corps_eval.calc당기순이익(nan_to_zero=True)

    c103.chg_page('c103현금흐름표q')
    d2, 재무활동현금흐름 = c103.sum_recent_4q(title='재무활동으로인한현금흐름', nan_to_zero=True)
    d3, 영업활동현금흐름 = c103.sum_recent_4q(title='영업활동으로인한현금흐름', nan_to_zero=True)

    c104.chg_page('c104q')
    d4, roic = c104.sum_recent_4q('ROIC')
    d5, roe = c104.latest_value('ROE')
    pcr_dict = c104.find('PCR')

    주주수익률 = float('nan') if marketcap == 0 else round((재무활동현금흐름 / marketcap * -100), 2)
    이익지표 = float('nan') if marketcap == 0 else round((지배주주당기순이익 - 영업활동현금흐름) / marketcap, 5)

    logger.info(f'{code} fcf_dict : {fcf_dict}')
    logger.info(f"{code} market_cap : {marketcap}")
    logger.info(f'{code} pfcf_dict : {pfcf_dict}')
    logger.info(f'{code} pcr_dict : {pcr_dict}')

    return {
        '주주수익률': 주주수익률,
        '이익지표': 이익지표,
        '투자수익률': {'ROIC': roic, 'ROE': roe},
        '가치지표': {'FCF': fcf_dict, 'PFCF': pfcf_dict, 'PCR': pcr_dict},
        'date': [i for i in {d1, d2, d3, d4, d5} if i != ''],  # ''값을 제거하고 리스트로 바꾼다.
    }


def blue(code: str) -> dict:
    """
    <유동비율>
    100미만이면 주의하나 현금흐름창출력이 좋으면 괜찮을수 있다.
    만약 100%이하면 유동자산에 추정영업현금흐름을 더해서 다시계산해보아 기회를 준다.
    <이자보상배율>
    이자보상배율 영업이익/이자비용으로 1이면 자금사정빡빡 5이상이면 양호
    <순운전자금회전율>
    순운전자금 => 기업활동을 하기 위해 필요한 자금 (매출채권 + 재고자산 - 매입채무)
    순운전자본회전율은 매출액/순운전자본으로 일정비율이 유지되는것이 좋으며 너무 작아지면 순운전자본이 많아졌다는 의미로 재고나 외상이 쌓인다는 뜻
    <재고자산회전율>
    재고자산회전율은 매출액/재고자산으로 회전율이 낮을수록 재고가 많다는 이야기이므로 불리 전년도등과 비교해서 큰차이 발생하면 알람.
    재고자산회전율이 작아지면 재고가 쌓인다는뜻
    <순부채비율>
    부채비율은 업종마다 달라 일괄비교 어려우나 순부채 비율이 20%이하인것이 좋고 꾸준히 늘어나지 않는것이 좋다.
    순부채 비율이 30%이상이면 좋치 않다.
    <매출액>
    매출액은 어떤경우에도 성장하는 기업이 좋다.매출이 20%씩 늘어나는 종목은 유망한 종목
    <영업이익률>
    영업이익률은 기업의 경쟁력척도로 경쟁사에 비해 높으면 경제적해자를 갖춘셈
    """

    if utils.is_6digit(code):
        corps_eval = CorpsEval(code)
        c104 = mongo.C104(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    ###############################################################

    d1, 유동비율 = corps_eval.calc유동비율(nan_to_zero=False)
    logger.info(f'유동비율 {유동비율}/{d1}')

    ###############################################################

    c104.chg_page('c104y')
    dict이자보상배율 = c104.find('이자보상배율')
    dict순운전자본회전율 = c104.find('순운전자본회전율')
    dict재고자산회전율 = c104.find('재고자산회전율')
    dict순부채비율 = c104.find('순부채비율')

    c104.chg_page('c104q')
    d2, 이자보상배율 = c104.latest_value('이자보상배율')
    d3, 순운전자본회전율 = c104.latest_value('순운전자본회전율')
    d4, 재고자산회전율 = c104.latest_value('재고자산회전율')
    d5, 순부채비율 = c104.latest_value('순부채비율')

    logger.info(f'이자보상배율 : {이자보상배율} {dict이자보상배율}')
    logger.info(f'순운전자본회전율 : {순운전자본회전율} {dict순운전자본회전율}')
    logger.info(f'재고자산회전율 : {재고자산회전율} {dict재고자산회전율}')
    logger.info(f'순부채비율 : {순부채비율} {dict순부채비율}')

    ################################################################

    return {
        '유동비율': 유동비율,
        '이자보상배율': (이자보상배율, dict이자보상배율),
        '순운전자본회전율': (순운전자본회전율, dict순운전자본회전율),
        '재고자산회전율': (재고자산회전율, dict재고자산회전율),
        '순부채비율': (순부채비율, dict순부채비율),
        'date': [i for i in {d1, d2, d3, d4, d5} if i != ''],  # ''값을 제거하고 리스트로 바꾼다.
    }


def growth(code: str) -> dict:
    """
    <매출액>
    매출액은 어떤경우에도 성장하는 기업이 좋다.매출이 20%씩 늘어나는 종목은 유망한 종목
    <영업이익률>
    영업이익률은 기업의 경쟁력척도로 경쟁사에 비해 높으면 경제적해자를 갖춘셈
    """
    if utils.is_6digit(code):
        c104 = mongo.C104(code)
        c106 = mongo.C106(code)
    else:
        raise ValueError(f'Invalid code: {code}')

    c104.chg_page('c104y')
    dict매출액증가율 = c104.find('매출액증가율')

    c104.chg_page('c104q')
    d1, 매출액증가율 = c104.latest_value('매출액증가율')

    logger.info(f'매출액증가율 : {매출액증가율} {dict매출액증가율}')

    ################################################################

    # c106 에서 타 기업과 영업이익률 비교
    c106.chg_page('c106y')
    try:
        dict영업이익률 = c106.find('영업이익률')
    except KeyError:
        dict영업이익률 = {'Unnamed': float('nan')}
    logger.info(f'{code} 영업이익률 {dict영업이익률}')

    return {
        '매출액증가율': (매출액증가율, dict매출액증가율),
        '영업이익률': dict영업이익률,
        'date': [d1, ]}


"""
- 각분기의 합이 연이 아닌 타이틀(즉 sum_4q를 사용하면 안됨)
'*(지배)당기순이익'
'*(비지배)당기순이익'
'장기차입금'
'현금및예치금'
'매도가능금융자산'
'매도파생결합증권'
'만기보유금융자산'
'당기손익-공정가치측정금융부채'
'당기손익인식(지정)금융부채'
'단기매매금융자산'
'단기매매금융부채'
'예수부채'
'차입부채'
'기타부채'
'보험계약부채(책임준비금)'
'*CAPEX'
'ROE'
"""

"""
- sum_4q를 사용해도 되는 타이틀
'자산총계'
'당기순이익'
'유동자산'
'유동부채'
'비유동부채'

'영업활동으로인한현금흐름'
'재무활동으로인한현금흐름'
'ROIC'
"""


def _make_df_part(codes: list, q):
    def make_record(c: str) -> dict:
        # 장고에서 사용할 eval 테이블을 만들기 위해 각각의 레코드를 구성하는 함수
        c101 = mongo.C101(code=c).get_recent()

        red_dict = red(code=c)
        mil_dict = mil(code=c)
        growth_dict = growth(code=c)

        mil_date = mil_dict['date']
        red_date = red_dict['date']
        growth_date = growth_dict['date']

        return {
            'code': c101['코드'],
            '종목명': c101['종목명'],
            '주가': utils.to_int(c101['주가']),
            'PER': utils.to_float(c101['PER']),
            'PBR': utils.to_float(c101['PBR']),
            '시가총액': utils.to_float(c101['시가총액']),
            'RED': utils.to_int(red_dict['red_price']),
            '주주수익률': utils.to_float(mil_dict['주주수익률']),
            '이익지표': utils.to_float(mil_dict['이익지표']),
            'ROIC': utils.to_float(mil_dict['투자수익률']['ROIC']),
            'ROE': utils.to_float(mil_dict['투자수익률']['ROE']),
            'PFCF': utils.to_float(CorpsEval.get_recent(mil_dict['가치지표']['PFCF'])[1]),
            'PCR': utils.to_float(CorpsEval.get_recent(mil_dict['가치지표']['PCR'])[1]),
            '매출액증가율': utils.to_float(growth_dict['매출액증가율'][0]),
            'date': list(set(mil_date + red_date + growth_date))
        }

    t = len(codes)
    d = []
    for i, code in enumerate(codes):
        print(f'{i+1}/{t} {code}')
        try:
            d.append(make_record(code))
        except:
            logger.error(f'error on {code}')
            continue
    df = pd.DataFrame(d)
    logger.info(df)
    q.put(df)


def make_today_eval_df(refresh: bool = False) -> pd.DataFrame:
    """ 멀티프로세싱을 사용하여 전체 종목의 eval 을 데이터프레임으로 만들어 반환

    기본값으로 refresh 는 False 로 설정되어 당일자의 저장된 데이터프레임이 있으면 새로 생성하지 않고 mongo DB를 이용한다.

    Args:
        refresh (bool): 저장된 db 자료를 이용하지 않고 강제로 리프레시하는 경우
    """
    def _code_divider(entire_codes: list) -> tuple:
        # 전체 종목코드를 리스트로 넣으면 cpu 코어에 맞춰 나눠준다.
        # https://stackoverflow.com/questions/19086106/how-to-utilize-all-cores-with-python-multiprocessing
        def _split_list(alist, wanted_parts=1):
            # 멀티프로세싱할 갯수로 리스트를 나눈다.
            # https://www.it-swarm.dev/ko/python/%EB%8D%94-%EC%9E%91%EC%9D%80-%EB%AA%A9%EB%A1%9D%EC%9C%BC%EB%A1%9C-%EB%B6%84%ED%95%A0-%EB%B0%98%EC%9C%BC%EB%A1%9C-%EB%B6%84%ED%95%A0/957910776/
            length = len(alist)
            return [alist[i * length // wanted_parts: (i + 1) * length // wanted_parts]
                    for i in range(wanted_parts)]

        core = cpu_count()
        print(f'Get number of core for multiprocessing : {core}')
        n = core - 1
        if len(entire_codes) < n:
            n = len(entire_codes)
        print(f'Split total {len(entire_codes)} codes by {n} parts ...')
        divided_list = _split_list(entire_codes, wanted_parts=n)
        return n, divided_list

    today_str = datetime.datetime.today().strftime('%Y%m%d')
    df = mongo.EvalByDate(date=today_str).get_data()
    if refresh or len(df) == 0:
        print(setting.load())
        codes = mongo.Corps.get_all_corps()

        print('*' * 25, f"Eval all using multiprocess", '*' * 25)
        print(f'Total {len(codes)} items..')
        logger.info(codes)
        n, divided_list = _code_divider(codes)

        start_time = time.time()
        q = Queue()
        ths = []
        for i in range(n):
            ths.append(Process(target=_make_df_part, args=(divided_list[i], q)))
        for i in range(n):
            ths[i].start()
        df_list = []
        for i in range(n):
            df_list.append(q.get())
        # 부분데이터프레임들을 하나로 합침
        final_df = pd.concat(df_list, ignore_index=True)
        for i in range(n):
            ths[i].join()
        print(f'Total spent time : {round(time.time() - start_time, 2)} sec.')
        logger.debug(final_df)
        print(f"Save to mongo db(db: eval col: {today_str}")
        mongo.EvalByDate(date=today_str).save(final_df)
    else:
        print(f"Use saved dataframe from mongo db..")
        final_df = mongo.EvalByDate(date=today_str).get_data()
    return final_df


def yield_valid_spac() -> tuple:
    """
    전체 스팩주의 현재가를 평가하여 2000원 이하인 경우 yield한다.

    Returns:
        tuple: (code, name, price)
    """
    codes = mongo.Corps.get_all_corps()
    logger.info(f'len(codes) : {len(codes)}')
    print('<<< Finding valuable SPAC >>>')
    for i, code in enumerate(codes):
        name = krx.get_name(code)
        logger.info(f'code : {code} name : {name}')
        if '스팩' in str(name):
            logger.info(f'>>> spac - code : {code} name : {name}')
            price, _, _ = utils.get_price_now(code=code)
            if price <= 2000:
                logger.warning(f'현재가:{price}')
                print(f"code: {code} name: {name}, price: {price}")
                yield code, name, price



"""다양한 문자열 출력 형식에 맞춘 함수들
"""
from db_hj3415 import mongo
from .eval import red as eval_red, mil as eval_mil, blue as eval_blue, growth as eval_growth
from .score import red as score_red, mil as score_mil, blue as score_blue, growth as score_growth
from .db import CorpsEval
from util_hj3415 import utils
from krx_hj3415 import krx
import textwrap

import logging
logger = logging.getLogger(__name__)
formatter = logging.Formatter('%(levelname)s: [%(name)s] %(message)s')
ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.setLevel(logging.WARNING)


class MakeStr:
    
    separate_line = '\n' + ('-' * 65) + '\n'

    def __init__(self, code: str):
        self.code = code
        self.name = krx.get_name(code)

    def c101(self, full=True):
        c101 = mongo.C101(self.code).get_recent()
        logger.info(c101)

        title = '=' * 35 + f"\t{c101['코드']}\t\t{c101['종목명']}\t\t{c101['업종']}\t" + '=' * 35
        intro = textwrap.fill(f"{c101['intro']}", width=70)

        if full:
            price = (f"{c101['date']}\t\t"
                     f"주가: {utils.deco_num(c101['주가'])}원\t\t"
                     f"52주최고: {utils.deco_num(c101['최고52주'])}원\t"
                     f"52주최저: {utils.deco_num(c101['최저52주'])}원")
            info = (f"PER: {c101['PER']}\t\t"
                    f"PBR: {c101['PBR']}\t\t\t"
                    f"배당수익률: {c101['배당수익률']}%\t\t"
                    f"시가총액: {utils.get_kor_amount(utils.to_int(c101['시가총액']), omit='억')}\n"
                    f"업종PER: {c101['업종PER']}\t"
                    f"유통비율: {c101['유통비율']}%\t\t"
                    f"거래대금: {utils.to_억(c101['거래대금'])}원\t\t"
                    f"발행주식: {utils.to_만(c101['발행주식'])}주")
        else:
            price = (f"<< {c101['date']} >>\n"
                     f"주가: {utils.deco_num(c101['주가'])}원")
            info = (f"PER: {c101['PER']}\n"
                    f"업종PER: {c101['업종PER']}\n"
                    f"PBR: {c101['PBR']}\n"
                    f"배당수익률: {c101['배당수익률']}%\n"
                    f"유통비율: {c101['유통비율']}%\n"
                    f"발행주식: {utils.to_만(c101['발행주식'])}주\n"
                    f"시가총액: {utils.get_kor_amount(utils.to_int(c101['시가총액']), omit='억')}")

        return title + '\n' + intro + self.separate_line + price + '\n' + info

    def red(self, full=True) -> str:
        red_dict = eval_red(self.code)
        p, 괴리율 = score_red(self.code)
        logger.info(red_dict)

        title = f"Red\tPoint({p})\t괴리율({괴리율}%)\t{red_dict['date']}\n"
        if full:
            contents = (f"사업가치({utils.deco_num(red_dict['사업가치'])}억) "
                        f"+ 재산가치({utils.deco_num(red_dict['재산가치'])}억) "
                        f"- 부채({utils.deco_num(red_dict['부채평가'])}억) "
                        f"/ 발행주식({utils.to_만(red_dict['발행주식수'])}주) "
                        f"= {utils.deco_num(red_dict['red_price'])}원")
        else:
            contents = f"{utils.deco_num(red_dict['red_price'])}원"
        return title + contents

    def mil(self, full=True) -> str:
        mil_dict = eval_mil(self.code)
        p1, p2, p3, p4 = score_mil(self.code)
        logger.info(mil_dict)

        title = f"Millenial\tPoint({p1+p2+p3+p4})\t{mil_dict['date']}\n"
        if full:
            contents = (f"1. 주주수익률({p1}): {mil_dict['주주수익률']} %\n"
                        f"2. 이익지표({p2}): {mil_dict['이익지표']}\n"
                        f"3. 투자수익률({p3}): ROIC 4분기합: {mil_dict['투자수익률']['ROIC']}%, "
                        f"최근 ROE: {mil_dict['투자수익률']['ROE']}%\n"
                        f"4. 가치지표\n"
                        f"\tFCF: {mil_dict['가치지표']['FCF']}\n"
                        f"\tPFCF({p4}) : {mil_dict['가치지표']['PFCF']}\n"
                        f"\tPCR: {mil_dict['가치지표']['PCR']}")
        else:
            contents = (f"1. 주주수익률({p1}): {mil_dict['주주수익률']} %\n"
                        f"2. 이익지표({p2}): {mil_dict['이익지표']}\n"
                        f"3. 투자수익률({p3}): ROIC 4분기합: {mil_dict['투자수익률']['ROIC']}%, "
                        f"최근 ROE: {mil_dict['투자수익률']['ROE']}%\n"
                        f"4. 가치지표\tPFCF({p4}) : {CorpsEval.get_recent(mil_dict['가치지표']['PFCF'])}")
        return title + contents

    def blue(self, full=True) -> str:
        blue_dict = eval_blue(self.code)
        p1, p2, p3, p4, p5 = score_blue(self.code)
        logger.info(blue_dict)

        title = f"Blue\tPoint({p1+p2+p3+p4+p5})\t{blue_dict['date']}\n"
        if full:
            contents = (f"1. 유동비율({p1}): {blue_dict['유동비율']}(100이하 위험)\n"
                        f"2. 이자보상배율({p2}): {blue_dict['이자보상배율']}(1이하 위험 5이상 양호)\n"
                        f"3. 순부채비율({p3}): {blue_dict['순부채비율']}(30이상 not good)\n"
                        f"4. 순운전자본회전율({p4}): {blue_dict['순운전자본회전율']}\n"
                        f"5. 재고자산회전율({p5}): {blue_dict['재고자산회전율']}")

        else:
            contents = ''
        return title + contents

    def growth(self, full=True) -> str:
        growth_dict = eval_growth(self.code)
        p1, p2 = score_growth(self.code)
        logger.info(growth_dict)

        title = f"Growth\tPoint({p1 + p2})\t{growth_dict['date']}\n"
        if full:
            contents = (f"1. 매출액증가율({p1}): {growth_dict['매출액증가율']}\n"
                        f"2. 영업이익률({p2}): {growth_dict['영업이익률']}")
        else:
            contents = (f"1. 매출액증가율({p1}): {growth_dict['매출액증가율'][0]}\n"
                        f"2. 영업이익률({p2}): {growth_dict['영업이익률'].get(self.name)}")
        return title + contents

    def c108(self, full=True) -> str:
        if full:
            c108_list = mongo.C108(self.code).get_all()
        else:
            c108_list = mongo.C108(self.code).get_recent()
        s = ''
        logger.info(c108_list)
        for i, c108_dict in enumerate(c108_list):
            logger.info(c108_dict)
            if i == 0:
                pass
            else:
                s += '\n'
            header = f"{c108_dict['날짜']}\thprice : {c108_dict['목표가']} 원\n"
            title = f"<<{c108_dict['제목']}>>\n"
            contents = ''
            for line in c108_dict['내용'].split('▶'):
                contents += line.strip()
            s += header + title + textwrap.fill(contents, width=70) + self.separate_line
        return s


def for_console(code: str) -> str:
    make_str = MakeStr(code=code)

    return (make_str.c101() + make_str.separate_line
            + make_str.red() + make_str.separate_line
            + make_str.mil() + make_str.separate_line
            + make_str.blue() + make_str.separate_line
            + make_str.growth() + make_str.separate_line
            + make_str.c108())


def for_telegram(code: str) -> str:
    make_str = MakeStr(code=code)

    return (make_str.c101(full=False) + make_str.separate_line
            + make_str.red(full=False) + make_str.separate_line
            + make_str.mil(full=False) + make_str.separate_line
            + make_str.blue(full=False) + make_str.separate_line
            + make_str.growth(full=False) + make_str.separate_line
            + make_str.c108(full=False))


def for_django(code: str) -> dict:
    """ 장고에서 report 페이지에서 사용될 eval data 를 반환

    장고의 view context는 딕셔너리 형식이기 때문에 딕셔너리 모음으로 반환한다.
    """
    return {
        'c101': mongo.C101(code).get_recent(),
        'red': eval_red(code),
        'mil': eval_mil(code),
        'blue': eval_blue(code),
        'growth': eval_growth(code),
        'c108': mongo.C108(code).get_recent(),
        'red_s': score_red(code),
        'mil_s': score_mil(code),
        'blue_s': score_blue(code),
        'growth_s': score_growth(code),
    }



import math
from db_hj3415.mongo import C103, C101, C104
from typing import Tuple, List
from util_hj3415 import utils
from collections import OrderedDict

import logging
logger = logging.getLogger(__name__)
formatter = logging.Formatter('%(levelname)s: [%(name)s] %(message)s')
ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.setLevel(logging.WARNING)


class CorpsEval:
    """eval에서 사용되는 여러가지 지수를 계산하는 함수 모음.

    """
    def __init__(self, code: str):
        if utils.is_6digit(code):
            self.code = code
            self.c101 = C101(code=code)
            self.c103 = C103(code=code)
            self.c104 = C104(code=code)
        else:
            raise ValueError(f'Invalid code : {code}')
        # C103이 억단위로 계산되기 때문에 시가총액도 억단위로 맞춘다.
        try:
            self.marketcap = int(self.c101.get_recent()['시가총액']) / 100000000
        except KeyError:
            self.marketcap = float('nan')

    def chg_code(self, code: str):
        if utils.is_6digit(code):
            self.code = code
            self.c101.chg_code(code=code)
            self.c103.chg_code(code=code)
            self.c104.chg_code(code=code)
        else:
            raise ValueError(f'Invalid code : {code}')
        # C103이 억단위로 계산되기 때문에 시가총액도 억단위로 맞춘다.
        try:
            self.marketcap = int(self.c101.get_recent()['시가총액']) / 100000000
        except KeyError:
            self.marketcap = float('nan')

    @staticmethod
    def sum_except_nan(targets: List[Tuple[str, float]]) -> Tuple[str, float]:
        """

        (date, value) 에서 value 가 nan 인 경우 해당 튜플을 건너뛰고 아닌 경우에만 합을 계산하여 구하는 함수

        Args:
            targets (List): (date, value) 형식의 튜플을 원소로 가지는 리스트

        Returns:
            Tuple: (date, 총합 value) 형식의 튜플
        """
        logger.info(f'sum_except_nan : {targets}')
        s_date = set()
        t_value = 0
        for date, value in targets:
            s_date.add(date)
            t_value += value
        try:
            return list(s_date)[0], round(t_value, 2)
        except IndexError:
            # s_date의 값이 하나도 없는 경우
            return '', 0

    def calc당기순이익(self, nan_to_zero: bool = True) -> Tuple[str, float]:
        """지배지분 당기순이익 계산

        일반적인 경우로는 직전 지배주주지분 당기순이익을 찾아서 반환한다.\n
        금융기관의 경우는 지배당기순이익이 없기 때문에\n
        계산을 통해서 간접적으로 구한다.\n

        Args:
            nan_to_zero (bool): 각 함수에서 반환하는 값이 nan 일 경우 0으로 치환할지 여부 결정

        """
        self.c103.chg_page('c103재무상태표q')
        profit_dict = self.c103.find(title='*(지배)당기순이익')
        logger.info(f'*(지배)당기순이익 : {profit_dict}')
        if len(profit_dict) == 0:
            # 금융관련은 재무상태표에 지배당기순이익이 없어서 손익계산서의 당기순이익에서 비지배당기순이익을 빼서 간접적으로 구한다.
            self.c103.chg_page('c103손익계산서q')
            t1 = self.c103.sum_recent_4q('당기순이익', nan_to_zero=nan_to_zero)
            logger.info(f'당기순이익(손익계산서) : {t1[0]}\t{t1[1]}')
            self.c103.chg_page('c103재무상태표q')
            t2 = self.c103.latest_value('*(비지배)당기순이익', nan_to_zero=nan_to_zero)
            logger.info(f'*(비지배)당기순이익 : {t2[0]}\t{t2[1]}')
            return self.sum_except_nan([t1, t2])
        else:
            return self.c103.latest_value('*(지배)당기순이익', nan_to_zero=nan_to_zero)

    def calc유동자산(self, nan_to_zero: bool = True) -> Tuple[str, float]:
        """유효한 유동자산 계산

        일반적인 경우로 유동자산을 찾아서 반환한다.\n
        금융기관의 경우는 간접적으로 계산한다.\n
        Red와 Blue에서 사용한다.\n

        Args:
            nan_to_zero (bool): 각 함수에서 반환하는 값이 nan 일 경우 0으로 치환할지 여부 결정

        """
        self.c103.chg_page('c103재무상태표q')
        asset_dict = self.c103.find(title='유동자산')
        logger.info(f'유동자산 : {asset_dict}')
        if len(asset_dict) == 0:
            # 금융관련업종...
            t1 = self.c103.latest_value('현금및예치금', nan_to_zero=nan_to_zero)
            logger.info(f'현금및예치금 : {t1[0]}\t{t1[1]}')
            t2 = self.c103.latest_value('단기매매금융자산', nan_to_zero=nan_to_zero)
            logger.info(f'단기매매금융자산 : {t2[0]}\t{t2[1]}')
            t3 = self.c103.latest_value('매도가능금융자산', nan_to_zero=nan_to_zero)
            logger.info(f'매도가능금융자산 : {t3[0]}\t{t3[1]}')
            t4 = self.c103.latest_value('만기보유금융자산', nan_to_zero=nan_to_zero)
            logger.info(f'만기보유금융자산 : {t4[0]}\t{t4[1]}')
            return self.sum_except_nan([t1, t2, t3, t4])
        else:
            return self.c103.sum_recent_4q('유동자산', nan_to_zero=nan_to_zero)

    def calc유동부채(self, nan_to_zero: bool = True) -> Tuple[str, float]:
        """유효한 유동부채 계산

        일반적인 경우로 유동부채를 찾아서 반환한다.\n
        금융기관의 경우는 간접적으로 계산한다.\n
        Red와 Blue에서 사용한다.\n

        Args:
            nan_to_zero (bool): 각 함수에서 반환하는 값이 nan 일 경우 0으로 치환할지 여부 결정

        """
        self.c103.chg_page('c103재무상태표q')
        debt_dict = self.c103.find(title='유동부채')
        logger.info(f'유동부채 : {debt_dict}')
        if len(debt_dict) == 0:
            # 금융관련업종...
            t1 = self.c103.latest_value('당기손익인식(지정)금융부채', nan_to_zero=nan_to_zero)
            logger.info(f'당기손익인식(지정)금융부채 : {t1[0]}\t{t1[1]}')
            t2 = self.c103.latest_value('당기손익-공정가치측정금융부채', nan_to_zero=nan_to_zero)
            logger.info(f'당기손익-공정가치측정금융부채 : {t2[0]}\t{t2[1]}')
            t3 = self.c103.latest_value('매도파생결합증권', nan_to_zero=nan_to_zero)
            logger.info(f'매도파생결합증권 : {t3[0]}\t{t3[1]}')
            t4 = self.c103.latest_value('단기매매금융부채', nan_to_zero=nan_to_zero)
            logger.info(f'단기매매금융부채 : {t4[0]}\t{t4[1]}')
            return self.sum_except_nan([t1, t2, t3, t4])
        else:
            return self.c103.sum_recent_4q('유동부채', nan_to_zero=nan_to_zero)

    def calc비유동부채(self, nan_to_zero: bool = True) -> Tuple[str, float]:
        """유효한 비유동부채 계산

        일반적인 경우로 비유동부채를 찾아서 반환한다.\n
        금융기관의 경우는 간접적으로 계산한다.\n
        Red와 Blue에서 사용한다.\n

        Args:
            nan_to_zero (bool): 각 함수에서 반환하는 값이 nan 일 경우 0으로 치환할지 여부 결정
        """
        self.c103.chg_page('c103재무상태표q')
        debt_dict = self.c103.find(title='비유동부채')
        logger.info(f'유동부채 : {debt_dict}')
        if len(debt_dict) == 0:
            # 금융관련업종...
            # 보험관련업종은 예수부채가 없는대신 보험계약부채가 있다...
            t1 = self.c103.latest_value('예수부채', nan_to_zero=nan_to_zero)
            logger.info(f'예수부채 : {t1[0]}\t{t1[1]}')
            t2 = self.c103.latest_value('보험계약부채(책임준비금)', nan_to_zero=nan_to_zero)
            logger.info(f'보험계약부채(책임준비금) : {t2[0]}\t{t2[1]}')
            t3 = self.c103.latest_value('차입부채', nan_to_zero=nan_to_zero)
            logger.info(f'차입부채 : {t3[0]}\t{t3[1]}')
            t4 = self.c103.latest_value('기타부채', nan_to_zero=nan_to_zero)
            logger.info(f'기타부채 : {t4[0]}\t{t4[1]}')
            return self.sum_except_nan([t1, t2, t3, t4])
        else:
            return self.c103.sum_recent_4q('비유동부채', nan_to_zero=nan_to_zero)

    def calc유동비율(self, nan_to_zero: bool = True) -> Tuple[str, float]:
        """유동비율계산 - Blue에서 사용

        c104q에서 최근유동비율 찾아보고 유효하지 않거나 \n
        100이하인 경우에는수동으로 계산해서 다시 한번 평가해 본다.\n

        Args:
            nan_to_zero (bool): 각 함수에서 반환하는 값이 nan 일 경우 0으로 치환할지 여부 결정
        """
        self.c104.chg_page('c104q')
        d1, 유동비율 = self.c104.latest_value('유동비율', nan_to_zero=False)
        logger.info(f'{self.code} 유동비율 : {유동비율}')

        if math.isnan(유동비율) or 유동비율 < 100:
            logger.warning('유동비율 is under 100 or nan..so we will recalculate..')
            d1, 유동자산 = self.calc유동자산(nan_to_zero=True)
            d2, 유동부채 = self.calc유동부채(nan_to_zero=True)

            self.c103.chg_page('c103현금흐름표q')
            d3, 추정영업현금흐름 = self.c103.sum_recent_4q('영업활동으로인한현금흐름', nan_to_zero=True)

            try:
                logger.info(f'{self.code} 계산전 유동비율 : {유동비율}')
                계산된유동비율 = round(((유동자산 + 추정영업현금흐름) / 유동부채) * 100, 2)
                logger.info(f'{self.code} 계산된 유동비율 : {계산된유동비율}')
                return list({d1, d2, d3})[0], 계산된유동비율
            except ZeroDivisionError:
                if nan_to_zero:
                    return list({d1, d2, d3})[0], 0
                else:
                    return list({d1, d2, d3})[0], float('nan')
        else:
            return d1, 유동비율

    def findFCF(self) -> dict:
        """FCF 계산

        FCF = 영업활동현금흐름 - CAPEX\n
        영업활동현금흐름에서 CAPEX 를 각 연도별로 빼주어 fcf 를 구하고 딕셔너리로 반환한다.\n

        Returns:
            dict: 계산된 fcf 딕셔너리 또는 영업현금흐름 없는 경우 - {}

        Note:
            CAPEX 가 없는 업종은 영업활동현금흐름을 그대로 사용한다.\n

        """
        self.c103.chg_page('c103현금흐름표y')
        영업활동으로인한현금흐름 = self.c103.find(title='영업활동으로인한현금흐름')
        self.c103.chg_page('c103재무상태표y')
        capex = self.c103.find(title='*CAPEX')

        logger.debug(f'영업활동으로인한현금흐름 {영업활동으로인한현금흐름}')
        logger.debug(f'CAPEX {capex}')

        if len(영업활동으로인한현금흐름) == 0:
            return {}

        if len(capex) == 0:
            # CAPEX 가 없는 업종은 영업활동현금흐름을 그대로 사용한다.
            return 영업활동으로인한현금흐름

        # 영업 활동으로 인한 현금 흐름에서 CAPEX 를 각 연도별로 빼주어 fcf 를 구하고 리턴값으로 fcf 딕셔너리를 반환한다.
        r_dict = {}
        for i in range(len(영업활동으로인한현금흐름)):
            # 영업활동현금흐름에서 아이템을 하나씩 꺼내서 CAPEX 전체와 비교하여 같으면 차를 구해서 r_dict 에 추가한다.
            date_cashflow, value_cashflow = 영업활동으로인한현금흐름.popitem()
            # 해당 연도의 capex 가 없는 경우도 있어 일단 capex를 0으로 치고 먼저 추가한다.
            r_dict[date_cashflow] = value_cashflow
            for date_capex, value_capex in capex.items():
                if date_cashflow == date_capex:
                    r_dict[date_cashflow] = round(value_cashflow - value_capex, 2)
        logger.debug(f'r_dict {r_dict}')
        # 연도순으로 정렬해서 딕셔너리로 반환한다.
        return dict(sorted(r_dict.items(), reverse=False))

    def findPFCF(self) -> dict:
        """Price to Free Cash Flow Ratio 계산

        PFCF = 시가총액 / FCF

        Note:
            https://www.investopedia.com/terms/p/pricetofreecashflow.asp
        """
        fcf_dict = self.findFCF()
        logger.debug(f'fcf_dict : {fcf_dict}')
        pfcf_dict = {}
        for date, fcf_one in fcf_dict.items():
            if fcf_one == 0:
                continue
            else:
                pfcf_dict[date] = round(self.marketcap / fcf_one, 2)
        logger.debug(f'pfcf_dict : {pfcf_dict}')
        return pfcf_dict

    @staticmethod
    def get_recent(data: dict, nan_to_zero: bool = False) -> Tuple[str, float]:
        """연/분기 딕셔너리에서 최근 자료를 추출해 반환함.

        최근 자료를 일차로 추출해서 nan 이면 한번 더 그 이전 자료를 추출해서 반환한다.

        Args:
             data(dict): 연/분기 재무재표 딕셔너리 자료
             nan_to_zero (bool): 각 함수에서 반환하는 값이 nan 일 경우 0으로 치환할지 여부 결정
        """
        # 딕셔너리 정렬 - https://kkamikoon.tistory.com/138
        # reverse = False 이면 오래된것부터 최근순으로 정렬한다.
        od = OrderedDict(sorted(data.items(), reverse=False))
        logger.info(f'{od}')
        try:
            last_one = od.popitem(last=True)
        except KeyError:
            # when dictionary is empty
            if nan_to_zero:
                return '', 0
            else:
                return '', float('nan')

        logger.info(f'last_one : {last_one}')
        if isinstance(last_one[1], str):
            # last_one : ('Unnamed: 1', '데이터가 없습니다.') 인 경우
            if nan_to_zero:
                return '', 0
            else:
                return '', float('nan')
        elif math.isnan(last_one[1]):
            try:
                last_one = od.popitem(last=True)
            except KeyError:
                # when dictionary is empty
                if nan_to_zero:
                    return '', 0
                else:
                    return '', float('nan')
        if nan_to_zero:
            return last_one[0], 0 if math.isnan(last_one[1]) else last_one[1]
        else:
            return tuple(last_one)
